@@ -15,13 +15,12 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import datetime
 import functools
 import json
 import warnings
-from typing import Any, Callable, TypeVar, Union, cast
+from typing import Any, Callable, TypeVar, cast
 
-import sqlalchemy
+import sqlalchemy as sa
 import sqlalchemy.orm
 
 from . import clean, config, database, decode, encode, utils
@@ -30,7 +29,7 @@ F = TypeVar("F", bound=Callable[..., Any])
 
 
 def _decode_and_update(
-    session: sqlalchemy.orm.Session,
+    session: sa.orm.Session,
     cache_entry: Any,
     settings: config.Settings,
 ) -> Any:
@@ -45,24 +44,6 @@ def _decode_and_update(
     return result
 
 
-def _delete_cache_entry(
-    session: sqlalchemy.orm.Session, cache_entry: database.CacheEntry
-) -> None:
-    # First, delete database entry
-    session.delete(cache_entry)
-    database._commit_or_rollback(session)
-    # Then, delete files
-    json.loads(cache_entry._result_as_string, object_hook=clean._delete_cache_file)
-
-
-def _hexdigestify_python_call(
-    func_to_hex: Union[str, Callable[..., Any]],
-    *args: Any,
-    **kwargs: Any,
-) -> str:
-    return utils.hexdigestify(encode.dumps_python_call(func_to_hex, *args, **kwargs))
-
-
 def cacheable(func: F) -> F:
     """Make a function cacheable."""
 
@@ -74,7 +55,7 @@ def cacheable(func: F) -> F:
             return func(*args, **kwargs)
 
         try:
-            hexdigest = _hexdigestify_python_call(func, *args, **kwargs)
+            hexdigest = encode._hexdigestify_python_call(func, *args, **kwargs)
         except encode.EncodeError as ex:
             if settings.return_cache_entry:
                 raise ex
@@ -83,15 +64,15 @@ def cacheable(func: F) -> F:
 
         filters = [
             database.CacheEntry.key == hexdigest,
-            database.CacheEntry.expiration > datetime.datetime.utcnow(),
+            database.CacheEntry.expiration > utils.utcnow(),
         ]
         if settings.expiration:
             # When expiration is provided, only get entries with matching expiration
             filters.append(database.CacheEntry.expiration == settings.expiration)
 
         with settings.sessionmaker() as session:
-            for cache_entry in (
-                session.query(database.CacheEntry)
+            for cache_entry in session.scalars(
+                sa.select(database.CacheEntry)
                 .filter(*filters)
                 .order_by(database.CacheEntry.timestamp.desc())
             ):
@@ -99,7 +80,7 @@ def cacheable(func: F) -> F:
                     return _decode_and_update(session, cache_entry, settings)
                 except decode.DecodeError as ex:
                     warnings.warn(str(ex), UserWarning)
-                    _delete_cache_entry(session, cache_entry)
+                    clean._delete_cache_entry(session, cache_entry)
 
         result = func(*args, **kwargs)
         cache_entry = database.CacheEntry(
