@@ -14,30 +14,29 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import collections
 import datetime
 import functools
 import json
 import posixpath
-from typing import Any, Dict, Literal, Optional, Sequence, Set
+from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Set, Union
 
-import sqlalchemy
+import fsspec
+import packaging.version
+import sqlalchemy as sa
 import sqlalchemy.orm
-import structlog
 
-from . import config, database, extra_encoders, utils
-
-LOGGER = structlog.get_logger()
+from . import config, database, decode, encode, extra_encoders, utils
 
 
 def _delete_cache_file(
     obj: Dict[str, Any],
-    session: Optional[sqlalchemy.orm.Session] = None,
+    session: Optional[sa.orm.Session] = None,
     cache_entry: Optional[database.CacheEntry] = None,
     sizes: Optional[Dict[str, int]] = None,
     dry_run: bool = False,
-    logger: Optional[structlog.stdlib.BoundLogger] = None,
 ) -> Any:
-    logger = logger or LOGGER
+    logger = config.get().logger
 
     if {"type", "callable", "args", "kwargs"} == set(obj) and obj["callable"] in (
         "cacholote.extra_encoders:decode_xr_dataset",
@@ -52,45 +51,92 @@ def _delete_cache_file(
 
         if posixpath.dirname(urlpath) == cache_dirname:
             sizes.pop(urlpath, None)
-            if session and cache_entry and not dry_run:
-                logger.info("Delete cache entry", cache_entry=cache_entry)
-                session.delete(cache_entry)
-                database._commit_or_rollback(session)
             if not dry_run:
-                with utils._Locker(fs, urlpath) as file_exists:
-                    if file_exists:
-                        logger.info("Delete cache file", urlpath=urlpath)
-                        fs.rm(urlpath, recursive=True)
+                if session and cache_entry:
+                    logger.info("delete cache entry", cache_entry=cache_entry)
+                    session.delete(cache_entry)
+                    database._commit_or_rollback(session)
+
+                if fs.exists(urlpath):
+                    logger.info("delete cache file", urlpath=urlpath)
+                    fs.rm(
+                        urlpath,
+                        recursive=obj["args"][0]["type"] == "application/vnd+zarr",
+                    )
 
     return obj
 
 
+def _delete_cache_entry(
+    session: sa.orm.Session, cache_entry: database.CacheEntry
+) -> None:
+    # First, delete database entry
+    session.delete(cache_entry)
+    database._commit_or_rollback(session)
+    # Then, delete files
+    json.loads(cache_entry._result_as_string, object_hook=_delete_cache_file)
+
+
+def delete(
+    func_to_del: Union[str, Callable[..., Any]], *args: Any, **kwargs: Any
+) -> None:
+    """Delete function previously cached.
+
+    Parameters
+    ----------
+    func_to_del: callable, str
+        Function to delete from cache
+    *args: Any
+        Argument of functions to delete from cache
+    **kwargs: Any
+        Keyword arguments of functions to delete from cache
+    """
+    hexdigest = encode._hexdigestify_python_call(func_to_del, *args, **kwargs)
+    with config.get().sessionmaker() as session:
+        for cache_entry in session.scalars(
+            sa.select(database.CacheEntry).filter(database.CacheEntry.key == hexdigest)
+        ):
+            _delete_cache_entry(session, cache_entry)
+
+
 class _Cleaner:
-    def __init__(self, logger: structlog.stdlib.BoundLogger) -> None:
-        self.logger = logger
+    def __init__(self) -> None:
+        self.logger = config.get().logger
+        self.fs, self.dirname = utils.get_cache_files_fs_dirname()
 
-        fs, dirname = utils.get_cache_files_fs_dirname()
-        sizes = {fs.unstrip_protocol(path): fs.du(path) for path in fs.ls(dirname)}
+        urldir = self.fs.unstrip_protocol(self.dirname)
 
-        self.fs = fs
-        self.dirname = dirname
-        self.sizes = sizes
+        self.logger.info("get disk usage of cache files")
+        self.sizes: Dict[str, int] = collections.defaultdict(lambda: 0)
+        for path, size in self.fs.du(self.dirname, total=False).items():
+            # Group dirs
+            urlpath = self.fs.unstrip_protocol(path)
+            basename, *_ = urlpath.replace(urldir, "", 1).strip("/").split("/")
+            if basename:
+                self.sizes[posixpath.join(urldir, basename)] += size
 
     @property
     def size(self) -> int:
-        return sum(self.sizes.values())
+        sum_sizes = sum(self.sizes.values())
+        self.logger.info("check cache files total size", size=sum_sizes)
+        return sum_sizes
 
     def stop_cleaning(self, maxsize: int) -> bool:
-        size = self.size
-        self.logger.info("Check cache files size", size=self.size)
-        return size <= maxsize
+        return self.size <= maxsize
 
     def get_unknown_files(self, lock_validity_period: Optional[float]) -> Set[str]:
-        now = datetime.datetime.now()
+        self.logger.info("get unknown files")
+
+        utcnow = (
+            utils.utcnow()
+            if packaging.version.parse(fsspec.__version__)
+            >= packaging.version.parse("2023.9.0")
+            else datetime.datetime.utcnow()
+        )
         files_to_skip = []
         for urlpath in self.sizes:
             if urlpath.endswith(".lock"):
-                delta = now - self.fs.modified(urlpath)
+                delta = utcnow - self.fs.modified(urlpath)
                 if lock_validity_period is None or delta < datetime.timedelta(
                     seconds=lock_validity_period
                 ):
@@ -100,28 +146,25 @@ class _Cleaner:
         unknown_sizes = {k: v for k, v in self.sizes.items() if k not in files_to_skip}
         if unknown_sizes:
             with config.get().sessionmaker() as session:
-                for cache_entry in session.query(database.CacheEntry):
+                for cache_entry in session.scalars(sa.select(database.CacheEntry)):
                     json.loads(
                         cache_entry._result_as_string,
                         object_hook=functools.partial(
                             _delete_cache_file,
                             sizes=unknown_sizes,
                             dry_run=True,
-                            logger=self.logger,
                         ),
                     )
         return set(unknown_sizes)
 
-    def delete_unknown_files(self, lock_validity_period: Optional[float]) -> None:
+    def delete_unknown_files(
+        self, lock_validity_period: Optional[float], recursive: bool
+    ) -> None:
         for urlpath in self.get_unknown_files(lock_validity_period):
             self.sizes.pop(urlpath)
-            if not self.fs.exists(urlpath):
-                continue
-
-            with utils._Locker(self.fs, urlpath, lock_validity_period) as file_exists:
-                if file_exists:
-                    self.logger.info("Delete unkown file", urlpath=urlpath)
-                    self.fs.rm(urlpath)
+            if self.fs.exists(urlpath):
+                self.logger.info("delete unknown", urlpath=urlpath, recursive=recursive)
+                self.fs.rm(urlpath, recursive=recursive)
 
     @staticmethod
     def check_tags(*args: Any) -> None:
@@ -149,7 +192,7 @@ class _Cleaner:
         filters = []
         if tags_to_keep is not None:
             filters.append(
-                sqlalchemy.or_(
+                sa.or_(
                     database.CacheEntry.tag.not_in(tags_to_keep),
                     database.CacheEntry.tag.is_not(None)
                     if None in tags_to_keep
@@ -158,7 +201,7 @@ class _Cleaner:
             )
         elif tags_to_clean is not None:
             filters.append(
-                sqlalchemy.or_(
+                sa.or_(
                     database.CacheEntry.tag.in_(tags_to_clean),
                     database.CacheEntry.tag.is_(None)
                     if None in tags_to_clean
@@ -167,20 +210,21 @@ class _Cleaner:
             )
 
         # Sorters
+        sorters: List[sa.orm.InstrumentedAttribute[Any]] = []
         if method == "LRU":
-            sorters = [database.CacheEntry.timestamp, database.CacheEntry.counter]
+            sorters.extend([database.CacheEntry.timestamp, database.CacheEntry.counter])
         elif method == "LFU":
-            sorters = [database.CacheEntry.counter, database.CacheEntry.timestamp]
+            sorters.extend([database.CacheEntry.counter, database.CacheEntry.timestamp])
         else:
-            raise ValueError("`method` must be 'LRU' or 'LFU'.")
+            raise ValueError(f"{method=} is invalid. Choose either 'LRU' or 'LFU'.")
         sorters.append(database.CacheEntry.expiration)
 
         # Clean database files
         if self.stop_cleaning(maxsize):
             return
         with config.get().sessionmaker() as session:
-            for cache_entry in (
-                session.query(database.CacheEntry).filter(*filters).order_by(*sorters)
+            for cache_entry in session.scalars(
+                sa.select(database.CacheEntry).filter(*filters).order_by(*sorters)
             ):
                 json.loads(
                     cache_entry._result_as_string,
@@ -189,7 +233,6 @@ class _Cleaner:
                         session=session,
                         cache_entry=cache_entry,
                         sizes=self.sizes,
-                        logger=self.logger,
                     ),
                 )
                 if self.stop_cleaning(maxsize):
@@ -204,10 +247,10 @@ def clean_cache_files(
     maxsize: int,
     method: Literal["LRU", "LFU"] = "LRU",
     delete_unknown_files: bool = False,
+    recursive: bool = False,
     lock_validity_period: Optional[float] = None,
     tags_to_clean: Optional[Sequence[Optional[str]]] = None,
     tags_to_keep: Optional[Sequence[Optional[str]]] = None,
-    logger: Optional[structlog.stdlib.BoundLogger] = None,
 ) -> None:
     """Clean cache files.
 
@@ -220,19 +263,19 @@ def clean_cache_files(
         * LFU: Least Frequently Used
     delete_unknown_files: bool, default: False
         Delete all files that are not registered in the cache database.
+    recursive: bool, default: False
+        Whether to delete unknown directories or not
     lock_validity_period: float, optional, default: None
         Validity period of lock files in seconds. Expired locks will be deleted.
     tags_to_clean, tags_to_keep: sequence of strings/None, optional, default: None
         Tags to clean/keep. If None, delete all cache entries.
         To delete/keep untagged entries, add None in the sequence (e.g., [None, 'tag1', ...]).
         tags_to_clean and tags_to_keep are mutually exclusive.
-    logger: optional
-        Python object use to produce logs.
     """
-    cleaner = _Cleaner(logger=logger or LOGGER)
+    cleaner = _Cleaner()
 
     if delete_unknown_files:
-        cleaner.delete_unknown_files(lock_validity_period)
+        cleaner.delete_unknown_files(lock_validity_period, recursive)
 
     cleaner.delete_cache_files(
         maxsize=maxsize,
@@ -240,3 +283,34 @@ def clean_cache_files(
         tags_to_clean=tags_to_clean,
         tags_to_keep=tags_to_keep,
     )
+
+
+def clean_invalid_cache_entries(
+    check_expiration: bool = True, try_decode: bool = False
+) -> None:
+    """Clean invalid cache entries.
+
+    Parameters
+    ----------
+    check_expiration: bool
+        Whether or not to delete expired entries
+    try_decode: bool
+        Whether or not to delete entries that raise DecodeError (this can be slow!)
+    """
+    filters = []
+    if check_expiration:
+        filters.append(database.CacheEntry.expiration <= utils.utcnow())
+    if filters:
+        with config.get().sessionmaker() as session:
+            for cache_entry in session.scalars(
+                sa.select(database.CacheEntry).filter(*filters)
+            ):
+                _delete_cache_entry(session, cache_entry)
+
+    if try_decode:
+        with config.get().sessionmaker() as session:
+            for cache_entry in session.scalars(sa.select(database.CacheEntry)):
+                try:
+                    decode.loads(cache_entry._result_as_string)
+                except decode.DecodeError:
+                    _delete_cache_entry(session, cache_entry)
