@@ -14,7 +14,7 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-
+import contextlib
 import functools
 import hashlib
 import inspect
@@ -23,7 +23,8 @@ import mimetypes
 import pathlib
 import posixpath
 import tempfile
-from typing import Any, Callable, Dict, Optional, Tuple, TypeVar, Union, cast
+import time
+from typing import Any, Callable, Dict, Generator, Optional, Tuple, TypeVar, Union, cast
 
 import fsspec
 import fsspec.implementations.local
@@ -68,6 +69,22 @@ def _add_ext_to_mimetypes() -> None:
 _add_ext_to_mimetypes()
 
 
+def _guess_type(
+    fs: fsspec.AbstractFileSystem,
+    local_path: str,
+    default: str = "application/octet-stream",
+) -> str:
+    content_type: str = fs.info(local_path).get("ContentType", "")
+    if content_type:
+        return content_type
+
+    filetype, *_ = mimetypes.guess_type(local_path, strict=False)
+    if filetype is None and _HAS_MAGIC:
+        with fs.open(local_path, "rb") as f:
+            filetype = magic.from_buffer(f.read(), mime=True)
+    return filetype or default
+
+
 def _requires_xarray_and_dask(func: F) -> F:
     """Raise an error if `xarray` or `dask` are not installed."""
 
@@ -80,36 +97,43 @@ def _requires_xarray_and_dask(func: F) -> F:
     return cast(F, wrapper)
 
 
+def _filesystem_is_local(fs: fsspec.AbstractFileSystem) -> bool:
+    return isinstance(fs, fsspec.get_filesystem_class("file"))
+
+
+@contextlib.contextmanager
+def _logging_timer(event: str, **kwargs: Any) -> Generator[float, None, None]:
+    logger = config.get().logger
+    logger.info(f"start {event}", **kwargs)
+    tic = time.perf_counter()
+    yield tic
+    toc = time.perf_counter()
+    kwargs["_".join(event.split() + ["time"])] = toc - tic  # elapsed time
+    logger.info(f"end {event}", **kwargs)
+
+
 class FileInfoModel(pydantic.BaseModel):
     type: str
     href: str
-    file_checksum: int = pydantic.Field(..., alias="file:checksum")
+    file_checksum: str = pydantic.Field(..., alias="file:checksum")
     file_size: int = pydantic.Field(..., alias="file:size")
     file_local_path: str = pydantic.Field(..., alias="file:local_path")
 
 
 def _dictify_file(fs: fsspec.AbstractFileSystem, local_path: str) -> Dict[str, Any]:
-    filetype = mimetypes.guess_type(local_path, strict=False)[0]
-    if filetype is None and _HAS_MAGIC:
-        with fs.open(local_path, "rb") as f:
-            filetype = magic.from_buffer(f.read(), mime=True)
-            if filetype == "application/octet-stream":
-                filetype = None
-    filetype = filetype or "unknown"
-
+    href = posixpath.join(
+        config.get().cache_files_urlpath_readonly or config.get().cache_files_urlpath,
+        posixpath.basename(local_path),
+    )
     file_dict = {
-        "type": filetype,
-        "href": posixpath.join(
-            config.get().cache_files_urlpath_readonly
-            or config.get().cache_files_urlpath,
-            posixpath.basename(local_path),
-        ),
-        "file:checksum": fs.checksum(local_path),
+        "type": _guess_type(fs, local_path),
+        "href": href,
+        "file:checksum": f"{fs.checksum(local_path):x}",
         "file:size": fs.size(local_path),
         "file:local_path": local_path,
     }
 
-    return FileInfoModel(**file_dict).dict(by_alias=True)
+    return FileInfoModel(**file_dict).model_dump(by_alias=True)
 
 
 def _get_fs_and_urlpath(
@@ -122,23 +146,32 @@ def _get_fs_and_urlpath(
         storage_options = config.get().cache_files_storage_options
 
     if not validate:
-        fs, _, _ = fsspec.get_fs_token_paths(urlpath, storage_options=storage_options)
+        fs, *_ = fsspec.get_fs_token_paths(urlpath, storage_options=storage_options)
         return (fs, urlpath)
 
     # Attempt to read from local_path
     try:
-        fs, _, _ = fsspec.get_fs_token_paths(urlpath, storage_options=storage_options)
+        fs, *_ = fsspec.get_fs_token_paths(urlpath, storage_options=storage_options)
     except:  # noqa: E722
         pass
     else:
         if fs.exists(urlpath):
-            if fs.checksum(urlpath) == file_json["file:checksum"]:
-                return (fs, urlpath)
-            raise ValueError("checksum mismatch")
+            expected = file_json["file:checksum"]
+            actual = (
+                fs.checksum(urlpath)  # Just for backward compatibility.
+                if isinstance(expected, int)
+                else f"{fs.checksum(urlpath):x}"
+            )
+            if expected != actual:
+                raise ValueError(f"checksum mismatch: {urlpath=} {expected=} {actual=}")
+            config.get().logger.info(
+                "retrieve cache file", urlpath=fs.unstrip_protocol(urlpath)
+            )
+            return (fs, urlpath)
 
     # Attempt to read from href
     urlpath = file_json["href"]
-    fs, _, _ = fsspec.get_fs_token_paths(urlpath)
+    fs, *_ = fsspec.get_fs_token_paths(urlpath)
     if fs.exists(urlpath):
         return (fs, urlpath)
 
@@ -159,15 +192,15 @@ def decode_xr_dataset(
     if file_json["type"] == "application/vnd+zarr":
         filename_or_obj = fs.get_mapper(urlpath)
     else:
-        if fsspec.utils.get_protocol(urlpath) == "file":
+        if _filesystem_is_local(fs):
             filename_or_obj = urlpath
         else:
             # Download local copy
-            protocol = fsspec.utils.get_protocol(urlpath)
+            protocols = (fs.protocol,) if isinstance(fs.protocol, str) else fs.protocol
             with fsspec.open(
                 f"filecache::{urlpath}",
                 filecache={"same_names": True},
-                **{protocol: fs.storage_options},
+                **{protocol: fs.storage_options for protocol in protocols},
             ) as of:
                 filename_or_obj = of.name
     return xr.open_dataset(filename_or_obj, **kwargs)
@@ -186,36 +219,38 @@ def decode_io_object(
 def _maybe_store_xr_dataset(
     obj: "xr.Dataset", fs: fsspec.AbstractFileSystem, urlpath: str, filetype: str
 ) -> None:
-    with utils._Locker(fs, urlpath) as file_exists:
-        if not file_exists:
-            if filetype == "application/vnd+zarr":
+    if filetype == "application/vnd+zarr":
+        with utils.FileLock(
+            fs, urlpath, timeout=config.get().lock_timeout
+        ) as file_exists:
+            if not file_exists:
                 # Write directly on any filesystem
                 mapper = fs.get_mapper(urlpath)
-                obj.to_zarr(mapper, consolidated=True)
-            else:
-                # Need a tmp local copy to write on a different filesystem
-                with tempfile.TemporaryDirectory() as tmpdirname:
-                    tmpfilename = str(
-                        pathlib.Path(tmpdirname) / pathlib.Path(urlpath).name
-                    )
+                with _logging_timer("upload", urlpath=fs.unstrip_protocol(urlpath)):
+                    obj.to_zarr(mapper, consolidated=True)
+    elif not fs.exists(urlpath):
+        # Need a tmp local copy to write on a different filesystem
+        with tempfile.TemporaryDirectory() as tmpdirname:
+            tmpfilename = str(pathlib.Path(tmpdirname) / pathlib.Path(urlpath).name)
 
-                    if filetype == "application/netcdf":
-                        obj.to_netcdf(tmpfilename)
-                    elif filetype == "application/x-grib":
-                        import cfgrib.xarray_to_grib
+            with _logging_timer("write tmp file", urlpath=tmpfilename):
+                if filetype == "application/netcdf":
+                    obj.to_netcdf(tmpfilename)
+                elif filetype == "application/x-grib":
+                    import cfgrib.xarray_to_grib
 
-                        cfgrib.xarray_to_grib.to_grib(obj, tmpfilename)
-                    else:
-                        # Should never get here! xarray_cache_type is checked in config.py
-                        raise ValueError(f"type {filetype!r} is NOT supported.")
+                    cfgrib.xarray_to_grib.to_grib(obj, tmpfilename)
+                else:
+                    # Should never get here! xarray_cache_type is checked in config.py
+                    raise ValueError(f"type {filetype!r} is NOT supported.")
 
-                    if fs == fsspec.filesystem("file"):
-                        fsspec.filesystem("file").mv(tmpfilename, urlpath)
-                    else:
-                        with fsspec.open(tmpfilename, "rb") as f_in, fs.open(
-                            urlpath, "wb"
-                        ) as f_out:
-                            utils.copy_buffered_file(f_in, f_out)
+            _maybe_store_file_object(
+                fs if _filesystem_is_local(fs) else fsspec.filesystem("file"),
+                tmpfilename,
+                fs,
+                urlpath,
+                io_delete_original=True,
+            )
 
 
 @_requires_xarray_and_dask
@@ -228,7 +263,7 @@ def dictify_xr_dataset(obj: "xr.Dataset") -> Dict[str, Any]:
     ext = mimetypes.guess_extension(filetype, strict=False)
     urlpath_out = posixpath.join(config.get().cache_files_urlpath, f"{root}{ext}")
 
-    fs_out, _, _ = fsspec.get_fs_token_paths(
+    fs_out, *_ = fsspec.get_fs_token_paths(
         urlpath_out,
         storage_options=config.get().cache_files_storage_options,
     )
@@ -252,21 +287,41 @@ def _maybe_store_file_object(
     urlpath_in: str,
     fs_out: fsspec.AbstractFileSystem,
     urlpath_out: str,
+    io_delete_original: Optional[bool] = None,
 ) -> None:
-    with utils._Locker(fs_out, urlpath_out) as file_exists:
+    if io_delete_original is None:
+        io_delete_original = config.get().io_delete_original
+    with utils.FileLock(
+        fs_out, urlpath_out, timeout=config.get().lock_timeout
+    ) as file_exists:
         if not file_exists:
-            if fs_in == fs_out:
-                if config.get().io_delete_original:
-                    fs_in.mv(urlpath_in, urlpath_out)
+            kwargs = {}
+            content_type = _guess_type(fs_in, urlpath_in)
+            if content_type:
+                kwargs["ContentType"] = content_type
+            with _logging_timer(
+                "upload",
+                urlpath=fs_out.unstrip_protocol(urlpath_out),
+                size=fs_in.size(urlpath_in),
+            ):
+                if fs_in == fs_out:
+                    if io_delete_original:
+                        fs_in.mv(urlpath_in, urlpath_out, **kwargs)
+                    else:
+                        fs_in.cp(urlpath_in, urlpath_out, **kwargs)
+                elif _filesystem_is_local(fs_in):
+                    fs_out.put(urlpath_in, urlpath_out, **kwargs)
                 else:
-                    fs_in.cp(urlpath_in, urlpath_out)
-            else:
-                with fs_in.open(urlpath_in, "rb") as f_in, fs_out.open(
-                    urlpath_out, "wb"
-                ) as f_out:
-                    utils.copy_buffered_file(f_in, f_out)
-                if config.get().io_delete_original and fs_in.exists(urlpath_in):
-                    fs_in.rm(urlpath_in)
+                    with fs_in.open(urlpath_in, "rb") as f_in:
+                        with fs_out.open(urlpath_out, "wb") as f_out:
+                            utils.copy_buffered_file(f_in, f_out)
+    if io_delete_original and fs_in.exists(urlpath_in):
+        with _logging_timer(
+            "remove",
+            urlpath=fs_in.unstrip_protocol(urlpath_in),
+            size=fs_in.size(urlpath_in),
+        ):
+            fs_in.rm(urlpath_in)
 
 
 def _maybe_store_io_object(
@@ -274,16 +329,19 @@ def _maybe_store_io_object(
     fs_out: fsspec.AbstractFileSystem,
     urlpath_out: str,
 ) -> None:
-    with utils._Locker(fs_out, urlpath_out) as file_exists:
+    with utils.FileLock(
+        fs_out, urlpath_out, timeout=config.get().lock_timeout
+    ) as file_exists:
         if not file_exists:
             f_out = fs_out.open(urlpath_out, "wb")
-            utils.copy_buffered_file(f_in, f_out)
+            with _logging_timer("upload", urlpath=fs_out.unstrip_protocol(urlpath_out)):
+                utils.copy_buffered_file(f_in, f_out)
 
 
 def dictify_io_object(obj: _UNION_IO_TYPES) -> Dict[str, Any]:
     """Encode a file object to JSON deserialized data (``dict``)."""
     cache_files_urlpath = config.get().cache_files_urlpath
-    fs_out, _, _ = fsspec.get_fs_token_paths(
+    fs_out, *_ = fsspec.get_fs_token_paths(
         cache_files_urlpath,
         storage_options=config.get().cache_files_storage_options,
     )
