@@ -1,11 +1,16 @@
+import contextlib
+import datetime
 import os
 import pathlib
-from typing import Any, Dict
+from typing import Any, Dict, Union
 
 import pytest
+import sqlalchemy as sa
 
 from cacholote import config
 
+does_not_raise = contextlib.nullcontext
+
 
 def test_change_cache_db_urlpath(tmpdir: pathlib.Path) -> None:
     old_db = config.get().cache_db_urlpath
@@ -81,3 +86,33 @@ def test_env_variables(tmpdir: pathlib.Path) -> None:
     finally:
         os.environ.clear()
         os.environ.update(old_environ)
+
+
+@pytest.mark.parametrize("poolclass", ("NullPool", sa.pool.NullPool))
+def test_set_poolclass(poolclass: Union[str, sa.pool.Pool]) -> None:
+    config.set(create_engine_kwargs={"poolclass": poolclass})
+    settings = config.get()
+    assert settings.create_engine_kwargs["poolclass"] == sa.pool.NullPool
+    assert isinstance(settings.engine.pool, sa.pool.NullPool)
+
+
+@pytest.mark.parametrize(
+    "expiration,raises",
+    [
+        (
+            datetime.datetime.now(),
+            pytest.raises(ValueError, match="Expiration is missing the timezone info."),
+        ),
+        (
+            datetime.datetime.now().isoformat(),
+            pytest.raises(ValueError, match="Expiration is missing the timezone info."),
+        ),
+        (datetime.datetime.now(tz=datetime.timezone.utc), does_not_raise()),
+        (datetime.datetime.now(tz=datetime.timezone.utc).isoformat(), does_not_raise()),
+    ],
+)
+def test_set_expiration(
+    expiration: Union[datetime.datetime, str], raises: contextlib.nullcontext  # type: ignore[type-arg]
+) -> None:
+    with raises:
+        config.set(expiration=expiration)
