@@ -2,6 +2,7 @@ import pathlib
 
 import fsspec
 import pytest
+import structlog
 
 from cacholote import cache, config, decode, encode, extra_encoders, utils
 
@@ -45,7 +46,7 @@ def test_dictify_xr_dataset(tmpdir: pathlib.Path) -> None:
             {
                 "type": "application/netcdf",
                 "href": href,
-                "file:checksum": fsspec.filesystem("file").checksum(local_path),
+                "file:checksum": f"{fsspec.filesystem('file').checksum(local_path):x}",
                 "file:size": 669,
                 "file:local_path": local_path,
             },
@@ -123,6 +124,7 @@ def test_xr_corrupted_files(
     importorskip: str,
 ) -> None:
     pytest.importorskip(importorskip)
+    import dask
 
     config.set(xarray_cache_type=xarray_cache_type)
 
@@ -132,17 +134,56 @@ def test_xr_corrupted_files(
     cfunc = cache.cacheable(get_grib_ds)
     cfunc()
 
+    # Get cached file path
+    with dask.config.set({"tokenize.ensure-deterministic": True}):
+        root = dask.base.tokenize(expected)  # type: ignore[no-untyped-call]
+    cached_path = f"{dirname}/{root}{ext}"
+    assert fs.exists(cached_path)
+
     # Warn if file is corrupted
-    fs.touch(f"{dirname}/b8094ae0691cfa42c9b839679e162e3d{ext}", truncate=False)
-    touched_info = fs.info(f"{dirname}/b8094ae0691cfa42c9b839679e162e3d{ext}")
+    fs.touch(cached_path, truncate=False)
+    touched_info = fs.info(cached_path)
     with pytest.warns(UserWarning, match="checksum mismatch"):
         actual = cfunc()
     xr.testing.assert_identical(actual, expected)
-    assert fs.info(f"{dirname}/b8094ae0691cfa42c9b839679e162e3d{ext}") != touched_info
+    assert fs.info(cached_path) != touched_info
 
     # Warn if file is deleted
-    fs.rm(f"{dirname}/b8094ae0691cfa42c9b839679e162e3d{ext}", recursive=True)
+    fs.rm(cached_path, recursive=True)
     with pytest.warns(UserWarning, match="No such file or directory"):
         actual = cfunc()
     xr.testing.assert_identical(actual, expected)
-    assert fs.exists(f"{dirname}/b8094ae0691cfa42c9b839679e162e3d{ext}")
+    assert fs.exists(cached_path)
+
+
+def test_xr_logging(capsys: pytest.CaptureFixture[str]) -> None:
+    config.set(logger=structlog.get_logger(), raise_all_encoding_errors=True)
+
+    # Cache dataset
+    cfunc = cache.cacheable(get_grib_ds)
+    cached_ds = cfunc()
+    captured = iter(capsys.readouterr().out.splitlines())
+
+    line = next(captured)
+    assert "start write tmp file" in line
+    assert "urlpath=" in line
+
+    line = next(captured)
+    assert "end write tmp file" in line
+    assert "urlpath=" in line
+    assert "write_tmp_file_time=" in line
+
+    line = next(captured)
+    assert "start upload" in line
+    assert f"urlpath=file://{cached_ds.encoding['source']}" in line
+    assert "size=22597" in line
+
+    line = next(captured)
+    assert "end upload" in line
+    assert f"urlpath=file://{cached_ds.encoding['source']}" in line
+    assert "upload_time=" in line
+    assert "size=22597" in line
+
+    line = next(captured)
+    assert "retrieve cache file" in line
+    assert f"urlpath=file://{cached_ds.encoding['source']}" in line
