@@ -1,19 +1,22 @@
+import contextlib
 import hashlib
 import importlib
 import io
 import pathlib
-import threading
-from typing import Any, Dict, Tuple, Union
+import subprocess
+from typing import Any, Dict, Optional, Tuple, Union
 
 import fsspec
 import pytest
 import pytest_httpserver
+import structlog
 
 from cacholote import cache, config, decode, encode, extra_encoders, utils
 
 
-def open_url(url: str) -> fsspec.spec.AbstractBufferedFile:
-    with fsspec.open(url) as f:
+@cache.cacheable
+def cached_open(*args: Any, **kwargs: Any) -> fsspec.spec.AbstractBufferedFile:
+    with fsspec.open(*args, **kwargs) as f:
         return f
 
 
@@ -35,7 +38,6 @@ def test_dictify_io_object(tmpdir: pathlib.Path, io_delete_original: bool) -> No
     actual = extra_encoders.dictify_io_object(open(tmpfile, "rb"))
     href = f"{readonly_dir}/{tmp_hash}.txt"
     local_path = f"{tmpdir}/cache_files/{tmp_hash}.txt"
-    checksum = fsspec.filesystem("file").checksum(local_path)
     expected = {
         "type": "python_call",
         "callable": "cacholote.extra_encoders:decode_io_object",
@@ -43,7 +45,7 @@ def test_dictify_io_object(tmpdir: pathlib.Path, io_delete_original: bool) -> No
             {
                 "type": "text/plain",
                 "href": href,
-                "file:checksum": checksum,
+                "file:checksum": f"{fsspec.filesystem('file').checksum(local_path):x}",
                 "file:size": 4,
                 "file:local_path": local_path,
             },
@@ -66,12 +68,16 @@ def test_dictify_bytes_io_object(
     actual = extra_encoders.dictify_io_object(obj)["args"]
     obj_hash = hashlib.md5(f"{hash(obj)}".encode()).hexdigest()
     local_path = f"{tmpdir}/cache_files/{obj_hash}"
-    checksum = fsspec.filesystem("file").checksum(local_path)
+    type = (
+        "text/plain"
+        if importlib.util.find_spec("magic")
+        else "application/octet-stream"
+    )
     expected: Tuple[Dict[str, Any], ...] = (
         {
-            "type": "text/plain" if importlib.util.find_spec("magic") else "unknown",
+            "type": type,
             "href": local_path,
-            "file:checksum": checksum,
+            "file:checksum": f"{fsspec.filesystem('file').checksum(local_path):x}",
             "file:size": 4,
             "file:local_path": local_path,
         },
@@ -98,9 +104,8 @@ def test_copy_from_http_to_cache(
 
     # cache http file
     fs, dirname = utils.get_cache_files_fs_dirname()
-    cfunc = cache.cacheable(open_url)
     for expected_counter in (1, 2):
-        result = cfunc(url)
+        result = cached_open(url)
 
         # Check hits
         cur.execute("SELECT counter FROM cache_entries", ())
@@ -124,48 +129,93 @@ def test_io_corrupted_files(
 
     # cache file
     fs, dirname = utils.get_cache_files_fs_dirname()
-    cfunc = cache.cacheable(open_url)
-    cfunc(url)
+    cached_open(url)
 
     # Warn if file is corrupted
     fs.touch(f"{dirname}/{cached_basename}")
     touched_info = fs.info(f"{dirname}/{cached_basename}")
     with pytest.warns(UserWarning, match="checksum mismatch"):
-        result = cfunc(url)
+        result = cached_open(url)
     assert result.read() == b"test"
     assert fs.info(f"{dirname}/{cached_basename}") != touched_info
 
     # Warn if file is deleted
     fs.rm(f"{dirname}/{cached_basename}")
     with pytest.warns(UserWarning, match="No such file or directory"):
-        result = cfunc(url)
+        result = cached_open(url)
     assert result.read() == b"test"
     assert fs.exists(f"{dirname}/{cached_basename}")
 
 
-def test_io_locker_warning(tmpdir: pathlib.Path) -> None:
-    cfunc = cache.cacheable(open)
-
+@pytest.mark.parametrize(
+    "lock_timeout, raises_or_warns",
+    (
+        [None, pytest.warns(UserWarning, match="is locked")],
+        [0, pytest.raises(TimeoutError, match="is locked")],
+    ),
+)
+def test_io_locker(
+    tmpdir: pathlib.Path,
+    lock_timeout: Optional[float],
+    raises_or_warns: contextlib.nullcontext,  # type: ignore[type-arg]
+) -> None:
+    config.set(lock_timeout=lock_timeout, raise_all_encoding_errors=True)
     # Create tmpfile
     tmpfile = tmpdir / "test.txt"
     fsspec.filesystem("file").touch(tmpfile)
 
     # Acquire lock
     fs, dirname = utils.get_cache_files_fs_dirname()
-    file_hash = f"{fsspec.filesystem('file').checksum(tmpfile):x}"
-    lock = f"{dirname}/{file_hash}.txt.lock"
-    fs.touch(lock)
+    file_path = f"{dirname}/{fsspec.filesystem('file').checksum(tmpfile):x}.txt"
+    fs.touch(f"{file_path}.lock")
 
-    def release_lock(fs: fsspec.AbstractFileSystem, lock: str) -> None:
-        fs.rm(lock)
+    process = subprocess.Popen(f"sleep 0.1; rm {file_path}.lock", shell=True)
+    with raises_or_warns:
+        cached_open(tmpfile)
+    assert process.wait() == 0
 
-    # Threading
-    t1 = threading.Timer(0, cfunc, args=(tmpfile,))
-    t2 = threading.Timer(0.1, release_lock, args=(fs, lock))
-    with pytest.warns(
-        UserWarning, match=f"can NOT proceed until file is released: {lock!r}."
-    ):
-        t1.start()
-        t2.start()
-        t1.join()
-        t2.join()
+
+@pytest.mark.parametrize("set_cache", ["cads"], indirect=True)
+def test_content_type(tmpdir: pathlib.Path, set_cache: str) -> None:
+    tmpfile = str(tmpdir / "test.grib")
+    fsspec.filesystem("file").touch(tmpfile)
+    fs, _ = utils.get_cache_files_fs_dirname()
+    cached_grib = cached_open(tmpfile)
+    assert fs.info(cached_grib)["ContentType"] == "application/x-grib"
+
+
+@pytest.mark.parametrize("set_cache", ["cads"], indirect=True)
+def test_io_logging(capsys: pytest.CaptureFixture[str], tmpdir: pathlib.Path) -> None:
+    config.set(logger=structlog.get_logger(), io_delete_original=True)
+
+    # Cache file
+    tmpfile = tmpdir / "test.txt"
+    fsspec.filesystem("file").touch(tmpfile)
+    cached_file = cached_open(tmpfile)
+    captured = iter(capsys.readouterr().out.splitlines())
+
+    line = next(captured)
+    assert "start upload" in line
+    assert f"urlpath=s3://{cached_file.path}" in line
+    assert "size=0" in line
+
+    line = next(captured)
+    assert "end upload" in line
+    assert f"urlpath=s3://{cached_file.path}" in line
+    assert "upload_time=" in line
+    assert "size=0" in line
+
+    line = next(captured)
+    assert "start remove" in line
+    assert f"urlpath=file://{tmpfile}" in line
+    assert "size=0" in line
+
+    line = next(captured)
+    assert "end remove" in line
+    assert f"urlpath=file://{tmpfile}" in line
+    assert "remove_time=" in line
+    assert "size=0" in line
+
+    line = next(captured)
+    assert "retrieve cache file" in line
+    assert f"urlpath=s3://{cached_file.path}" in line
